import pypulseq as pp
import pytest

# Gradient definitions used in tests
gx_trap = pp.make_trapezoid('x', area=1000, duration=1e-3)
gx_extended = pp.make_extended_trapezoid('x', amplitudes=[0, 100000, 0], times=[0, 1e-4, 2e-4])
gx_extended_delay = pp.make_extended_trapezoid('x', amplitudes=[0, 100000, 0], times=[1e-4, 2e-4, 3e-4])
gx_endshigh = pp.make_extended_trapezoid('x', amplitudes=[0, 100000, 100000], times=[0, 1e-4, 2e-4])
gx_startshigh = pp.make_extended_trapezoid('x', amplitudes=[100000, 100000, 0], times=[0, 1e-4, 2e-4])
gx_startshigh2 = pp.make_extended_trapezoid('x', amplitudes=[200000, 100000, 0], times=[0, 1e-4, 2e-4])
gx_allhigh = pp.make_extended_trapezoid('x', amplitudes=[100000, 100000, 100000], times=[0, 1e-4, 2e-4])
delay = pp.make_delay(1e-3)

## Test gradient continuity checks in add_block


def test_gradient_continuity1():
    # Trap followed by extended gradient: No error
    seq = pp.Sequence()
    seq.add_block(gx_trap)
    seq.add_block(gx_extended)
    seq.add_block(gx_trap)


def test_gradient_continuity2():
    # Trap followed by non-zero gradient
    with pytest.raises(RuntimeError):
        seq = pp.Sequence()
        seq.add_block(gx_trap)
        seq.add_block(gx_startshigh)  # raises


def test_gradient_continuity3():
    # Gradient starts at non-zero in first block
    with pytest.raises(RuntimeError):
        seq = pp.Sequence()
        seq.add_block(gx_startshigh)  # raises


def test_gradient_continuity4():
    # Gradient starts and ends at non-zero
    with pytest.raises(RuntimeError):
        seq = pp.Sequence()
        seq.add_block(delay)
        seq.add_block(gx_allhigh)


def test_gradient_continuity5():
    # Gradient starts at zero and has a delay: No error
    seq = pp.Sequence()
    seq.add_block(gx_extended_delay)


def test_gradient_continuity6():
    # Gradient starts at non-zero in other blocks
    with pytest.raises(RuntimeError):
        seq = pp.Sequence()
        seq.add_block(delay)
        seq.add_block(gx_startshigh)  # raises


def test_gradient_continuity7():
    # Gradient ends high and is followed by empty block
    with pytest.raises(RuntimeError):
        seq = pp.Sequence()
        seq.add_block(gx_endshigh)
        seq.add_block(delay)  # raises


def test_gradient_continuity8():
    # Gradient ends high and is followed by trapezoid
    with pytest.raises(RuntimeError):
        seq = pp.Sequence()
        seq.add_block(gx_endshigh)
        seq.add_block(gx_trap)  # raises


def test_gradient_continuity9():
    # Gradient ends high and is followed by connecting gradient: No error
    seq = pp.Sequence()
    seq.add_block(gx_endshigh)
    seq.add_block(gx_startshigh)


def test_gradient_continuity10():
    # Gradient in last block ends high: No error, this is caught by seq.write()
    seq = pp.Sequence()
    seq.add_block(gx_endshigh)


def test_gradient_continuity11():
    # Non-zero, but non-connecting gradients
    with pytest.raises(RuntimeError):
        seq = pp.Sequence()
        seq.add_block(gx_endshigh)
        seq.add_block(gx_startshigh2)


## Test gradient continuity checks in set_block


def test_gradient_continuity_setblock1():
    # Use set_block to insert gradient
    with pytest.raises(RuntimeError):
        seq = pp.Sequence()
        seq.add_block(delay)
        seq.add_block(delay)
        seq.add_block(delay)

        seq.set_block(1, gx_startshigh)


def test_gradient_continuity_setblock2():
    with pytest.raises(RuntimeError):
        seq = pp.Sequence()
        seq.add_block(delay)
        seq.add_block(delay)
        seq.add_block(delay)

        seq.set_block(2, gx_startshigh)


def test_gradient_continuity_setblock3():
    with pytest.raises(RuntimeError):
        seq = pp.Sequence()
        seq.add_block(delay)
        seq.add_block(delay)
        seq.add_block(delay)

        seq.set_block(3, gx_startshigh)


def test_gradient_continuity_setblock4():
    # Overwrite valid gradient with empty block
    with pytest.raises(RuntimeError):
        seq = pp.Sequence()
        seq.add_block(gx_endshigh)
        seq.add_block(gx_allhigh)
        seq.add_block(gx_startshigh)

        seq.set_block(2, delay)


def test_gradient_continuity_setblock5():
    # Overwrite valid gradient with gradient that is valid on one side
    with pytest.raises(RuntimeError):
        seq = pp.Sequence()
        seq.add_block(gx_endshigh)
        seq.add_block(gx_allhigh)
        seq.add_block(gx_startshigh)

        seq.set_block(2, gx_startshigh)


def test_gradient_continuity_setblock6():
    # Add new gradient with non-contiguous block numbering
    with pytest.raises(RuntimeError):
        seq = pp.Sequence()
        seq.add_block(gx_endshigh)
        seq.add_block(gx_allhigh)
        seq.add_block(gx_startshigh)

        seq.set_block(6, gx_startshigh)


def test_gradient_continuity_setblock7():
    # Valid sequence with non-contiguous block numbering
    seq = pp.Sequence()
    seq.set_block(10, gx_endshigh)
    seq.set_block(5, gx_allhigh)
    seq.set_block(7, gx_startshigh)

    assert list(seq.block_events.keys()) == [10, 5, 7]


# TODO: Add other block functionality tests



from copy import copy
from types import SimpleNamespace
from typing import Union

import numpy as np

from pypulseq import eps
from pypulseq.opts import Opts


def scale_grad(grad: SimpleNamespace, scale: float, system: Union[Opts, None] = None) -> SimpleNamespace:
    """
    Scales the gradient with the scalar.

    Parameters
    ----------
    grad : SimpleNamespace
        Gradient event to be scaled.
    scale : float
        Scaling factor.
    system : Opts, default=Opts()
        System limits.

    Returns
    -------
    grad : SimpleNamespace
        Scaled gradient.
    """
    # copy() to emulate pass-by-value; otherwise passed grad event is modified
    scaled_grad = copy(grad)
    if scaled_grad.type == 'trap':
        scaled_grad.amplitude = scaled_grad.amplitude * scale
        scaled_grad.flat_area = scaled_grad.flat_area * scale
        if system is not None:
            if abs(scaled_grad.amplitude) > system.max_grad:
                raise ValueError(
                    f'scale_grad: maximum amplitude exceeded {100 * abs(scaled_grad.amplitude) / system.max_grad} %'
                )
            if (
                abs(grad.amplitude) > eps
                and abs(scaled_grad.amplitude) / min(scaled_grad.rise_time, scaled_grad.fall_time) > system.max_slew
            ):
                raise ValueError(
                    'mr.scale_grad: maximum slew rate exceeded {100 * abs(scaled_grad.amplitude) / min(scaled_grad.rise_time, scaled_grad.fall_time) / system.max_slew} %'
                )

    else:
        scaled_grad.waveform = scaled_grad.waveform * scale
        scaled_grad.first = scaled_grad.first * scale
        scaled_grad.last = scaled_grad.last * scale
        if system is not None:
            if max(abs(scaled_grad.waveform)) > system.max_grad:
                raise ValueError(
                    f'scale_grad: maximum amplitude exceeded {100 * max(abs(scaled_grad.waveform)) / system.max_grad} %'
                )
            if max(abs(scaled_grad.waveform)) > eps:
                scaled_grad_max_abs_slew = max(abs(np.diff(scaled_grad.waveform) / np.diff(grad.tt)))
                if scaled_grad_max_abs_slew > system.max_slew:
                    raise ValueError(
                        f'scale_grad: maximum slew rate exceeded {100 * scaled_grad_max_abs_slew / system.max_slew} %'
                    )
    scaled_grad.area = scaled_grad.area * scale

    if hasattr(scaled_grad, 'id'):
        delattr(scaled_grad, 'id')

    return scaled_grad



import re
import warnings
from types import SimpleNamespace
from typing import Callable, Dict, List, Tuple, Union

import numpy as np

from pypulseq.calc_duration import calc_duration
from pypulseq.calc_rf_center import calc_rf_center
from pypulseq.compress_shape import compress_shape
from pypulseq.decompress_shape import decompress_shape
from pypulseq.event_lib import EventLibrary
from pypulseq.supported_labels_rf_use import get_supported_labels


def read(self, path: str, detect_rf_use: bool = False, remove_duplicates: bool = True) -> None:
    """
    Load sequence from file - read the given filename and load sequence data into sequence object.

    See also `pypulseq.Sequence.write_seq.write()`.

    Parameters
    ----------
    path : Path
        Path of sequence file to be read.
    detect_rf_use : bool, default=False
        Boolean flag to let the function infer the currently missing flags concerning the intended use of the RF pulses
        (excitation, refocusing, etc). These are important for the k-space trajectory calculation.
    remove_duplicates: bool, default=True
        Remove duplicate events from the sequence after reading

    Raises
    ------
    FileNotFoundError
        If no sequence file is found at `path`.
    RuntimeError
        If incompatible sequence files are attempted to be loaded.
    ValueError
        If unexpected sections are encountered when loading a sequence file.
    """
    try:
        input_file = open(path, 'r')
    except FileNotFoundError as e:
        raise FileNotFoundError(e) from e

    # Event libraries
    self.adc_library = EventLibrary()
    self.grad_library = EventLibrary()
    self.label_inc_library = EventLibrary()
    self.label_set_library = EventLibrary()
    self.rf_library = EventLibrary()
    self.shape_library = EventLibrary()
    self.trigger_library = EventLibrary()

    # Raster times
    self.grad_raster_time = self.system.grad_raster_time
    self.rf_raster_time = self.system.rf_raster_time

    self.block_events = {}
    self.definitions = {}
    self.extension_string_idx = []
    self.extension_numeric_idx = []

    file_version_combined = 0

    # Load data from file
    while True:
        section = __skip_comments(input_file)
        if section == -1:
            break
        if section == '[DEFINITIONS]':
            self.definitions = __read_definitions(input_file)

            # Gradient raster time
            if 'GradientRasterTime' in self.definitions:
                self.gradient_raster_time = self.definitions['GradientRasterTime']

            # Radio frequency raster time
            if 'RadiofrequencyRasterTime' in self.definitions:
                self.rf_raster_time = self.definitions['RadiofrequencyRasterTime']

            # ADC raster time
            if 'AdcRasterTime' in self.definitions:
                self.adc_raster_time = self.definitions['AdcRasterTime']

            # Block duration raster
            if 'BlockDurationRaster' in self.definitions:
                self.block_duration_raster = self.definitions['BlockDurationRaster']
            else:
                warnings.warn(f'No BlockDurationRaster found in file. Using default of {self.block_duration_raster}.')

        elif section == '[SIGNATURE]':
            temp_sign_defs = __read_definitions(input_file)
            if 'Type' in temp_sign_defs:
                self.signature_type = temp_sign_defs['Type']
            if 'Hash' in temp_sign_defs:
                self.signature_value = temp_sign_defs['Hash']
                self.signature_file = 'Text'
        elif section == '[VERSION]':
            file_version_major, file_version_minor, file_version_revision = __read_version(input_file)
            file_version_combined = 1000000 * file_version_major + 1000 * file_version_minor + file_version_revision

            # Check if file version is higher than version of the installed PyPulseq package
            package_version_combined = (
                1000000 * self.version_major + 1000 * self.version_minor + int(self.version_revision)
            )
            if file_version_combined > package_version_combined:
                warnings.warn(
                    f'File version {file_version_major}.{file_version_minor}.{file_version_revision} is higher than '
                    f'installed package version {self.version_major}.{self.version_minor}.{self.version_revision}. '
                    f'This may cause parsing errors.'
                )

            if file_version_combined < 1002000:
                raise RuntimeError(
                    f'Unsupported version {file_version_major}.{file_version_minor}.{file_version_revision}, only file '
                    f'format revision 1.2.0 and above are supported.'
                )

            if file_version_combined < 1003001:
                raise RuntimeError(
                    f'Loading older Pulseq format file (version '
                    f'{file_version_major}.{file_version_minor}.{file_version_revision}) some code may function not as '
                    f'expected'
                )

            if file_version_combined >= 1005000 and detect_rf_use:
                warnings.warn('Option detectRFuse is not supported for file format version 1.5.0 and above')
                detect_rf_use = False

        elif section == '[BLOCKS]':
            if file_version_major == 0:
                raise RuntimeError('Pulseq file MUST include [VERSION] section prior to [BLOCKS] section')
            result = __read_blocks(
                input_file,
                block_duration_raster=self.block_duration_raster,
                version_combined=file_version_combined,
            )
            self.block_events, self.block_durations, delay_ind_temp = result
        elif section == '[RF]':
            if file_version_combined >= 1005000:  # 1.5.x format
                self.rf_library = __read_events(
                    input_file,
                    (1, 1, 1, 1, 1e-6, 1e-6, 1, 1, 1, 1, np.nan),
                    event_library=self.rf_library,
                )
            elif file_version_combined >= 1004000:  # 1.4.x format
                self.rf_library = __read_events(
                    input_file,
                    (1, 1, 1, 1, 1e-6, 1, 1),
                    event_library=self.rf_library,
                )
            else:  # 1.3.x and below
                self.rf_library = __read_events(input_file, (1, 1, 1, 1e-6, 1, 1), event_library=self.rf_library)

        elif section == '[GRADIENTS]':
            if file_version_combined >= 1005000:  # 1.5.x format
                self.grad_library = __read_events(input_file, (1, 1, 1, 1, 1, 1e-6), 'g', self.grad_library)
            elif file_version_combined >= 1004000:  # 1.4.x format
                self.grad_library = __read_events(input_file, (1, 1, 1, 1e-6), 'g', self.grad_library)
            else:  # 1.3.x and below
                self.grad_library = __read_events(input_file, (1, 1, 1e-6), 'g', self.grad_library)
        elif section == '[TRAP]':
            self.grad_library = __read_events(input_file, (1, 1e-6, 1e-6, 1e-6, 1e-6), 't', self.grad_library)
        elif section == '[ADC]':
            if file_version_combined >= 1005000:  # 1.5.x format
                self.adc_library = __read_events(
                    input_file,
                    (1, 1e-9, 1e-6, 1, 1, 1, 1, 1),
                    event_library=self.adc_library,
                    append=self.system.adc_dead_time,
                )
            else:  # 1.4.x format and below
                self.adc_library = __read_events(
                    input_file, (1, 1e-9, 1e-6, 1, 1), event_library=self.adc_library, append=self.system.adc_dead_time
                )
        elif section == '[DELAYS]':
            if file_version_combined >= 1004000:
                raise RuntimeError('Pulseq file revision 1.4.0 and above MUST NOT contain [DELAYS] section')
            temp_delay_library = __read_events(input_file, (1e-6,))
        elif section == '[SHAPES]':
            self.shape_library = __read_shapes(input_file, file_version_major == 1 and file_version_minor < 4)
        elif section == '[EXTENSIONS]':
            self.extensions_library = __read_events(input_file)
        else:
            if section[:18] == 'extension TRIGGERS':
                extension_id = int(section[18:])
                self.set_extension_string_ID('TRIGGERS', extension_id)
                self.trigger_library = __read_events(input_file, (1, 1, 1e-6, 1e-6), event_library=self.trigger_library)
            elif section[:18] == 'extension LABELSET':
                extension_id = int(section[18:])
                self.set_extension_string_ID('LABELSET', extension_id)

                def l1(s):
                    return int(s)

                def l2(s):
                    return get_supported_labels().index(s) + 1

                self.label_set_library = __read_and_parse_events(input_file, l1, l2)
            elif section[:18] == 'extension LABELINC':
                extension_id = int(section[18:])
                self.set_extension_string_ID('LABELINC', extension_id)

                def l1(s):
                    return int(s)

                def l2(s):
                    return get_supported_labels().index(s) + 1

                self.label_inc_library = __read_and_parse_events(input_file, l1, l2)
            elif section[:16] == 'extension DELAYS':
                extension_id = int(section[16:])
                self.set_extension_string_ID('DELAYS', extension_id)
                self.soft_delay_library = __read_and_parse_events(
                    input_file, int, lambda ofs: 1e-6 * int(ofs), int, str
                )
                # Map numIDs to soft delay hints
                for d in self.soft_delay_library.data.values():
                    if d[3] not in self.soft_delay_hints:
                        self.soft_delay_hints[d[3]] = d[0]
            else:
                raise ValueError(f'Unknown section code: {section}')

    input_file.close()  # Close file

    if file_version_combined < 1002000:
        raise ValueError(
            f'Unsupported version {file_version_combined}, only file format revision 1.2.0 (1002000) and above '
            f'are supported.'
        )

    # Fix blocks, gradients and RF objects imported from older versions (< v1.4.0)
    if file_version_combined < 1004000:
        # Scan through RF objects
        self.rf_library.type = dict.fromkeys(self.rf_library.type.keys(), 'u')
        for i in self.rf_library.data:
            d = self.rf_library.data[i]
            rf = self.rf_from_lib_data((d[:3], 0, 0, d[3], 0, 0, d[4:6], 'u')).__delattr__('center')
            center = calc_rf_center(rf)
            self.rf_library.update(
                i,
                None,
                (d[:3], 0, center, d[3], 0, 0, d[4:6], 'u'),
            )  # 0 between [3] and [4:6] are the freq_ppm and phase_ppm

        # Scan through the gradient objects and update 't'-s (trapezoids) und 'g'-s (free-shape gradients)
        for i in self.grad_library.data:
            if self.grad_library.type[i] == 't':
                if self.grad_library.data[i][1] == 0:  # noqa: SIM102
                    if abs(self.grad_library.data[i][0]) == 0 and self.grad_library.data[i][2] > 0:
                        d = self.grad_library.data[i]
                        self.grad_library.update(
                            i,
                            None,
                            (d[0], self.grad_raster_time, d[2] - self.grad_raster_time, *d[3:]),
                            self.grad_library.type[i],
                        )

                if self.grad_library.data[i][3] == 0:  # noqa: SIM102
                    if abs(self.grad_library.data[i][0]) == 0 and self.grad_library.data[i][2] > 0:
                        d = self.grad_library.data[i]
                        self.grad_library.update(
                            i,
                            None,
                            (*d[:2], d[2] - self.grad_raster_time, self.grad_raster_time, *d[4:]),
                            self.grad_library.type[i],
                        )

            if self.grad_library.type[i] == 'g':
                self.grad_library.update(
                    i,
                    None,
                    (
                        self.grad_library.data[i][:2],
                        0,
                        self.grad_library.data[i][2:],
                    ),
                    self.grad_library.type[i],
                )

        # For versions prior to 1.4.0 block_durations have not been initialized
        self.block_durations = {}
        # Scan through blocks and calculate durations
        for block_counter in self.block_events:
            # Insert delay as temporary block_duration
            self.block_durations[block_counter] = 0
            if delay_ind_temp[block_counter] > 0:
                self.block_durations[block_counter] = temp_delay_library.data[delay_ind_temp[block_counter]][0]

            block = self.get_block(block_counter)
            # Calculate actual block duration
            self.block_durations[block_counter] = calc_duration(block)

    elif file_version_combined < 1005000:
        # Port from v1.4.x : RF, ADC and GRAD objects need to be updated
        # this needs to be done on the level of the libraries, because get_block will fail

        # Scan though the RFs and add center, freq_ppm, phase_ppm and use fields
        self.rf_library.type = dict.fromkeys(self.rf_library.type.keys(), 'u')
        for i in self.rf_library.data:
            # Use goes into the type field, and this is done separately
            d = self.rf_library.data[i]
            rf = self.rf_from_lib_data((d[:4], 0, d[4], 0, 0, d[5:7], 'u')).__delattr__('center')
            center = calc_rf_center(rf)
            self.rf_library.update(
                i,
                None,
                (d[:4], center, d[4], 0, 0, d[5:7], 'u'),
            )  # 0 between [4] and [5:7] are the freqPPM and phasePPM

        # Scan through the gradient objects and update 'g'-s (free-shape gradients)
        for i in self.grad_library.data:
            if self.grad_library.type[i] == 'g':
                self.grad_library.update(
                    i,
                    None,
                    (
                        self.grad_library.data[i][0],
                        None,
                        None,
                        self.grad_library.data[i][1:4],
                    ),
                    self.grad_library.type[i],
                )  # We use None to label the non-initialized first/last fields. These will be restored in the code below

    # Another run through for all older versions
    if file_version_combined < 1005000:
        # TODO: Is it possible to avoid expensive get_block calls here?
        grad_channels = ['gx', 'gy', 'gz']
        grad_prev_last = np.zeros(len(grad_channels))
        for block_counter in self.block_events:
            block = self.get_block(block_counter)
            block_duration = block.block_duration
            # We also need to keep track of the event IDs because some PyPulseq files written by external software may contain
            # repeated entries so searching by content will fail
            event_idx = self.block_events[block_counter]
            # Update the objects by filling in the 'first' and 'last' attributes not yet contained in the Pulseq file
            for j in range(len(grad_channels)):
                grad = getattr(block, grad_channels[j])
                if grad is None:
                    grad_prev_last[j] = 0
                    continue

                if grad.type == 'grad':
                    if grad.delay > 0:
                        grad_prev_last[j] = 0

                    # go to next channel, if grad.first and grad.last are already set
                    if hasattr(grad, 'first') and hasattr(grad, 'last'):
                        grad_prev_last[j] = grad.last
                        continue

                    # get grad.first and grad.last attributes from the grad_library if they have been set for the current amplitude_ID before
                    amplitude_ID = event_idx[j + 2]
                    if amplitude_ID in event_idx[2 : (j + 2)]:
                        if self.use_block_cache:
                            grad.first = self.grad_library.data[amplitude_ID][4]
                            grad.last = self.grad_library.data[amplitude_ID][5]
                        continue

                    # get time_id from grad_library
                    time_id = self.grad_library.data[amplitude_ID][2]

                    # if grad.first is not set, set it to the last value of the previous gradient
                    grad.first = grad_prev_last[j]

                    # extended trapezoid: use last value of the gradient waveform as grad.last
                    if time_id != 0:
                        grad.last = grad.waveform[-1]
                        grad_duration = grad.delay + grad.tt[-1]
                    # arbitrary gradients: interpolate grad.last from the gradient waveform
                    else:
                        # use a linear extrapolation identical to the one used in the make_arbitrary_grad.py file
                        grad.last = (3 * grad.waveform[-1] - grad.waveform[-2]) * 0.5
                        grad_duration = grad.delay + len(grad.waveform) * self.grad_raster_time

                    # Set grad_prev_last to 0 if gradient does not end at block boundary
                    eps = np.finfo(np.float64).eps
                    if grad_duration + eps < block_duration:
                        grad_prev_last[j] = 0
                    # Update grad_prev_last for the next iteration if gradient ends at block boundary
                    else:
                        grad_prev_last[j] = grad.last

                    # Update the grad_library with the new grad.first and grad.last values
                    amplitude = self.grad_library.data[amplitude_ID][0]
                    shape_id = self.grad_library.data[amplitude_ID][1]
                    new_data = (
                        amplitude,
                        shape_id,
                        time_id,
                        grad.delay,
                        grad.first,
                        grad.last,
                    )
                    self.grad_library.update_data(amplitude_ID, None, new_data, 'g')

                else:
                    grad_prev_last[j] = 0

    if detect_rf_use:
        # Find the RF pulses, list flip angles, and work around the current (rev 1.2.0) Pulseq file format limitation
        # that the RF pulse use is not stored in the file
        for k in self.rf_library.data:
            lib_data = self.rf_library.data[k]
            rf = self.rf_from_lib_data(lib_data)
            flip_deg = np.abs(np.sum(rf.signal[:-1] * (rf.t[1:] - rf.t[:-1]))) * 360
            offresonance_ppm = 1e6 * rf.freq_offset / self.system.B0 / self.system.gamma
            if flip_deg < 90.01:  # Add 0.01 degree to account for rounding errors encountered in very short RF pulses
                self.rf_library.type[k] = 'e'
            else:
                if rf.shape_dur > 6e-3 and -3.5 <= offresonance_ppm <= -3.4:  # Approx -3.45
                    self.rf_library.type[k] = 's'  # Saturation (fat-sat)
                else:
                    self.rf_library.type[k] = 'r'
            self.rf_library.data[k] = lib_data

            # Clear block cache for all blocks that contain the modified RF event
            for block_counter, events in self.block_events.items():
                if events[1] == k:
                    del self.block_cache[block_counter]

    # When removing duplicates, remove and remap events in the sequence without
    # creating a copy.
    if remove_duplicates:
        self.remove_duplicates(in_place=True)


def __read_definitions(input_file) -> Dict[str, str]:
    """
    Read the [DEFINITIONS] section of a sequence fil and return a map of key/value entries.

    Parameters
    ----------
    input_file : file object
        Sequence file.

    Returns
    -------
    definitions : dict{str, str}
        Dict object containing key value pairs of definitions.
    """
    definitions = {}
    line = __skip_comments(input_file)
    while line != -1 and not (line == '' or line[0] == '#'):
        tok = line.split(' ')
        try:  # Try converting every element into a float
            [float(x) for x in tok[1:]]
            value = np.array(tok[1:], dtype=float)
            if len(value) == 1:  # Avoid array structure for single elements
                value = value[0]
            definitions[tok[0]] = value
        except ValueError:  # Try clause did not work!
            definitions[tok[0]] = line[len(tok[0]) + 1 :].strip()
        line = __strip_line(input_file)

    return definitions


def __read_version(input_file) -> Tuple[int, int, int]:
    """
     Read the [VERSION] section of a sequence file.

    Parameters
    ----------
    input_file : file object
        Sequence file.

    Returns
    -------
    tuple
        Major, minor and revision number.
    """
    line = __strip_line(input_file)
    major, minor, revision = 0, 0, 0
    while line != '' and line[0] != '#':
        tok = line.split(' ')
        if tok[0] == 'major':
            major = int(tok[1])
        elif tok[0] == 'minor':
            minor = int(tok[1])
        elif tok[0] == 'revision':
            if len(tok[1]) != 1:  # Example: x.y.zpostN
                tok[1] = tok[1][0]
            revision = int(tok[1])
        else:
            raise RuntimeError(f'Incompatible version. Expected: {major}{minor}{revision}')
        line = __strip_line(input_file)

    return major, minor, revision


def __read_blocks(
    input_file, block_duration_raster: float, version_combined: int
) -> Tuple[Dict[int, np.ndarray], List[float], List[int]]:
    """
    Read the [BLOCKS] section of a sequence file and return the event table.

    Parameters
    ----------
    input_file : file
        Sequence file

    Returns
    -------
    event_table : dict
        Dict object containing key value pairs of Pulseq block ID and block definition.
    block_durations : list
        Block durations.
    delay_idx : list
        Delay IDs.
    """
    event_table = {}
    block_durations = {}
    delay_idx = {}
    line = __strip_line(input_file)

    while line != '' and line != '#':
        block_events = np.fromstring(line, dtype=int, sep=' ')

        if version_combined <= 1002001:
            event_table[block_events[0]] = np.array([0, *block_events[2:], 0])
        else:
            event_table[block_events[0]] = np.array([0, *block_events[2:]])

        delay_id = block_events[0]
        if version_combined >= 1004000:
            block_durations[delay_id] = block_events[1] * block_duration_raster
        else:
            delay_idx[delay_id] = block_events[1]

        line = __strip_line(input_file)

    return event_table, block_durations, delay_idx


def __read_events(
    input_file,
    scale: Union[Tuple, None] = None,
    event_type: str = str(),
    event_library: EventLibrary = None,
    append=None,
) -> EventLibrary:
    """
    Read an event section of a sequence file and return a library of events.

    Parameters
    ----------
    input_file : file object
        Sequence file.
    scale : list, default=(1,)
        Scale elements according to column vector scale.
    event_type : str, default=str()
        Attach the type string to elements of the library.
    event_library : EventLibrary, default=EventLibrary()
        Append new events to the given library.

    Returns
    -------
    event_library : EventLibrary
        Event library containing Pulseq events.
    """
    if event_library is None:
        event_library = EventLibrary()

    # New in v1.5.0 : generate format string; NaN labels character param(s) for 'use' attribute
    line, scale, format_spec = __read_format(input_file, scale)
    data_mask = np.isfinite(scale)
    type_idx = np.where(np.logical_not(data_mask))[0]

    if len(type_idx) > 1:
        raise ValueError('Only one type field (marked as NaN) can be provided in scale.')
    if len(type_idx) == 0:
        type_idx = None
        data_mask = None
    else:
        type_idx = type_idx.item()

    while line != '' and line != '#':
        data = __fromstring(line, format_spec)
        event_id = data[0]

        if type_idx is not None:
            event_type = data[type_idx + 1]  # Need +1 because of the event_id in the first position

        data = data[1:]
        data = tuple(data[n] * scale[n] if isinstance(data[n], str) is False else data[n] for n in range(len(data)))

        if append is not None:
            data = (*data, append)

        if event_type == '' and type_idx is None:
            if data_mask is None:
                event_library.insert(key_id=event_id, new_data=data)
            else:
                event_library.insert(key_id=event_id, new_data=data[data_mask])
        else:
            if data_mask is None:
                event_library.insert(key_id=event_id, new_data=data, data_type=event_type)
            else:
                data = tuple(np.asarray(data, dtype=object)[data_mask])
                event_library.insert(key_id=event_id, new_data=data, data_type=event_type)

        line = __strip_line(input_file)

    return event_library


def __read_and_parse_events(input_file, *args: Callable) -> EventLibrary:
    """
    Read an event section of a sequence file and return a library of events.

    Event data elements are converted using the provided parser(s). Default parser is `int()`.

    Parameters
    ----------
    input_file : file
    args : Callable
        Event parsers.

    Returns
    -------
    EventLibrary
        Library of events parsed from the events section of a sequence file.
    """
    event_library = EventLibrary()
    line = __strip_line(input_file)

    while line != '' and line != '#':
        list_of_data_str = re.split(r'(\s+)', line)
        list_of_data_str = [d for d in list_of_data_str if d != ' ']
        data = []  # np.zeros(len(list_of_data_str) - 1, dtype=np.int32)
        event_id = int(list_of_data_str[0])
        for i in range(1, len(list_of_data_str)):
            if i > len(args):
                data.append(int(list_of_data_str[i]))
            else:
                data.append(args[i - 1](list_of_data_str[i]))
        event_library.insert(key_id=event_id, new_data=data)
        line = __strip_line(input_file)

    return event_library


def __read_shapes(input_file, force_convert_uncompressed: bool) -> EventLibrary:
    """
    Read the [SHAPES] section of a sequence file and return a library of shapes.

    Parameters
    ----------
    input_file : file

    Returns
    -------
    shape_library : EventLibrary
        `EventLibrary` object containing shape definitions.
    """
    shape_library = EventLibrary(numpy_data=True)

    line = __skip_comments(input_file)

    while line != -1 and (line != '' or line[0:8] == 'shape_id'):
        tok = line.split(' ')
        shape_id = int(tok[1])
        line = __skip_comments(input_file)
        tok = line.split(' ')
        num_samples = int(tok[1])
        data = []
        line = __skip_comments(input_file)
        while line != '' and line != '#':
            data.append(float(line))
            line = __strip_line(input_file)
        line = __skip_comments(input_file, stop_before_section=True)

        # Check if conversion is needed: in v1.4.x we use length(data)==num_samples
        # As a marker for the uncompressed (stored) data. In older versions this condition could occur by chance
        if force_convert_uncompressed and len(data) == num_samples:
            shape = SimpleNamespace()
            shape.data = data
            shape.num_samples = num_samples
            shape = compress_shape(decompress_shape(shape, force_decompression=True))
            data = np.array([shape.num_samples, *shape.data])
        else:
            data.insert(0, num_samples)
            data = np.asarray(data)
        shape_library.insert(key_id=shape_id, new_data=data)
    return shape_library


def __skip_comments(input_file, stop_before_section: bool = False) -> str:
    """
    Read lines of skipping blank lines and comments and return the next non-comment line.

    Parameters
    ----------
    input_file : file

    Returns
    -------
    line : str
        First line in `input_file` after skipping one '#' comment block. Note: File pointer is remembered, so
        successive calls work as expected.
    """
    temp_pos = input_file.tell()
    line = __strip_line(input_file)
    while line != -1 and (line == '' or line[0] == '#'):
        temp_pos = input_file.tell()
        line = __strip_line(input_file)

    if line != -1:
        if stop_before_section and line[0] == '[':
            input_file.seek(temp_pos, 0)
            next_line = ''
        else:
            next_line = line
    else:
        next_line = -1

    return next_line


def __strip_line(input_file) -> str:
    """
    Removes spaces and newline whitespaces.

    Parameters
    ----------
    input_file : file

    Returns
    -------
    line : str
        First line in input_file after spaces and newline whitespaces have been removed. Note: File pointer is
        remembered, and hence successive calls work as expected. Returns -1 for eof.
    """
    line = input_file.readline()  # If line is an empty string, end of the file has been reached
    return line.strip() if line != '' else -1


def __read_format(input_file, scale: Tuple) -> Tuple[List, Tuple, List]:
    """
    Generate a format specifier list based on the scale vector.

    '%f' for numeric values, '%s' for string (NaN in scale).

    Parameters
    ----------
    input_file: file
        Input text
    scale : list, default=(1,)
        Scale elements according to column vector scale.

    Returns
    -------
    line : str
        First line in input_file after spaces and newline whitespaces have been removed. Note: File pointer is
        remembered, and hence successive calls work as expected. Returns -1 for eof.
    scale : tuple[float]
        Scaling factor for each element in input line.
        Defaults to one for each numeric element.
    format : list[str]
        List of format tokens for each field (excluding the event ID).

    """
    line = __strip_line(input_file)

    if scale is None:
        tok = line.strip().split()
        scale = (len(tok) - 1) * (1,)

    scale = np.array(scale, dtype=float)
    is_num = np.isfinite(scale)
    format_spec = ['%f' if value else '%s' for value in is_num]

    return line, scale, format_spec


def __fromstring(line, format_spec: List) -> List:
    """
    Parse a line of text using a dynamic format specification.

    Parameters
    ----------
    line : str
        Input line (e.g., ['23', '1.0', '2.0', '3.0', 'u'])
    format_spec : list[str]
        Format list like ['%f', '%f', '%f', '%f', '%s']

    Returns
    -------
    data : list
        Parsed data.

    """
    tok = line.strip().split()
    if len(tok) != len(format_spec) + 1:
        raise ValueError('Mismatch between number of tokens and format spec')

    format_spec = ['%f', *format_spec]
    data = []

    for i, fmt in enumerate(format_spec):
        if fmt == '%f':
            data.append(float(tok[i]))
        elif fmt == '%s':
            data.append(tok[i])
        else:
            raise ValueError(f'Unsupported format: {fmt}')

    return data



from types import SimpleNamespace
from typing import Union

from pypulseq.opts import Opts


def make_digital_output_pulse(
    channel: str,
    delay: float = 0.0,
    duration: float = 4e-3,
    system: Union[Opts, None] = None,
) -> SimpleNamespace:
    """
    Create a digital output pulse event a.k.a. trigger. Creates an output trigger event on a given channel with optional
    given delay and duration.

    Parameters
    ----------
    channel : str
        Must be one of 'osc0','osc1', or 'ext1'.
    delay : float, default=0
        Delay in seconds (s).
    duration : float, default=4e-3
        Duration of trigger event in seconds (s).
    system : Opts, default=Opts()
        System limits.

    Returns
    -------
    trig : SimpleNamespace
        Trigger event.

    Raises
    ------
    ValueError
        If `channel` is invalid. Must be one of 'osc0','osc1', or 'ext1'.
    """
    if system is None:
        system = Opts.default

    if channel not in ['osc0', 'osc1', 'ext1']:
        raise ValueError(f"Channel {channel} is invalid. Must be one of 'osc0','osc1', or 'ext1'.")

    trig = SimpleNamespace()
    trig.type = 'output'
    trig.channel = channel
    trig.delay = delay
    trig.duration = duration
    if trig.duration <= system.grad_raster_time:
        trig.duration = system.grad_raster_time

    return trig



from types import SimpleNamespace
from typing import Union

import numpy as np

from pypulseq import eps
from pypulseq.opts import Opts
from pypulseq.utils.tracing import trace, trace_enabled


def make_arbitrary_grad(
    channel: str,
    waveform: np.ndarray,
    first: Union[float, None] = None,
    last: Union[float, None] = None,
    delay: float = 0.0,
    max_grad: Union[float, None] = None,
    max_slew: Union[float, None] = None,
    system: Union[Opts, None] = None,
    oversampling: bool = False,
) -> SimpleNamespace:
    """
    Creates a gradient event from an arbitrary waveform.

    Note that the sample points are assumed to be equally spaced by `system.grad_raster_time`
    and that the given waveform values are the values in the middle of each raster interval.

    The duration of the gradient is thus given by the number of samples times `system.grad_raster_time`.

    See also `pypulseq.Sequence.sequence.Sequence.add_block()`.

    Parameters
    ----------
    channel : str
        Orientation of gradient event of arbitrary shape. Must be one of `x`, `y` or `z`.
    waveform : numpy.ndarray
        Arbitrary waveform.
    first : float
        Gradient value at the start of the gradient event. (t=0)
        Will default to a linear extrapolated value if not provided.
    last : float
        Gradient value at the end of the gradient event. (t=duration)
        Will default to a linear extrapolated value if not provided.
    system : Opts, default=Opts()
        System limits.
        Will default to `pypulseq.opts.default` if not provided.
    max_grad : float
        Maximum gradient strength.
        Will default to `system.max_grad` if not provided.
    max_slew : float
        Maximum slew rate.
        Will default to `system.max_slew` if not provided.
    delay : float, default=0
        Delay in seconds (s).
    oversampling : bool, default=False
        Boolean flag to indicate if gradient is oversampled by a factor of 2.

    Returns
    -------
    grad : SimpleNamespace
        Gradient event with arbitrary waveform.

    Raises
    ------
    ValueError
        If invalid `channel` is passed. Must be one of x, y or z.
        If slew rate is violated.
        If gradient amplitude is violated.
    """
    if system is None:
        system = Opts.default

    if max_grad is None or max_grad == 0:
        max_grad = system.max_grad

    if max_slew is None or max_slew == 0:
        max_slew = system.max_slew

    if channel not in ['x', 'y', 'z']:
        raise ValueError(f'Invalid channel. Must be one of x, y or z. Passed: {channel}')

    if first is None or last is None:

        def extrap(a, b):
            # extrapolate by 1 gradient raster (oversampling)
            # or by 1/2 gradient of the raster (non-oversampling)
            return 2 * a - b if oversampling else 0.5 * (3 * a - b)

        if first is None:
            first = extrap(waveform[0], waveform[1])
        if last is None:
            last = extrap(waveform[-1], waveform[-2])

    # Slew rate calculation
    if oversampling:
        edge_scale = system.grad_raster_time * 2
        pre = first - waveform[0]
        post = last - waveform[-1]
    else:
        edge_scale = system.grad_raster_time
        pre = 2 * (first - waveform[0])
        post = 2 * (waveform[-1] - last)

    slew_rate = np.concatenate([[pre], np.diff(waveform), [post]]) / edge_scale

    if max(abs(slew_rate)) > max_slew * (1 + eps):
        raise ValueError(f'Slew rate violation {max(abs(slew_rate)) / max_slew * 100}')
    if max(abs(waveform)) > max_grad + eps:
        raise ValueError(f'Gradient amplitude violation {max(abs(waveform)) / max_grad * 100}')

    grad = SimpleNamespace()
    grad.type = 'grad'
    grad.channel = channel
    grad.waveform = waveform
    grad.delay = delay
    if oversampling:
        if len(waveform) % 2 == 0:
            raise ValueError('When oversampling is active, waveform must have an odd number of samples')
        grad.area = (waveform[::2] * system.grad_raster_time).sum()
        grad.tt = np.arange(1, len(waveform) + 1) * 0.5 * system.grad_raster_time
        grad.shape_dur = (len(waveform) + 1) * 0.5 * system.grad_raster_time
    else:
        grad.area = (waveform * system.grad_raster_time).sum()
        grad.tt = (np.arange(len(waveform)) + 0.5) * system.grad_raster_time
        grad.shape_dur = len(waveform) * system.grad_raster_time
    grad.first = first
    grad.last = last

    if trace_enabled():
        grad.trace = trace()

    return grad



import numpy as np


def points_to_waveform(amplitudes: np.ndarray, grad_raster_time: float, times: np.ndarray) -> np.ndarray:
    """
    1D interpolate amplitude values `amplitudes` at time indices `times` as per the gradient raster time
    `grad_raster_time` to generate a gradient waveform.

    Parameters
    ----------
    amplitudes : numpy.ndarray
        Amplitude values at time indices `times`.
    grad_raster_time : float
        Gradient raster time.
    times : numpy.ndarray
        Time indices.

    Returns
    -------
    waveform : numpy.ndarray
        Gradient waveform.
    """
    amplitudes = np.asarray(amplitudes)
    times = np.asarray(times)

    if amplitudes.size == 0:
        return np.array([0])

    grd = (
        np.arange(
            start=round(np.min(times) / grad_raster_time),
            stop=round(np.max(times) / grad_raster_time),
        )
        * grad_raster_time
    )
    waveform = np.interp(x=grd + grad_raster_time / 2, xp=times, fp=amplitudes)

    return waveform



from copy import deepcopy
from types import SimpleNamespace
from typing import List, Union

import numpy as np

from pypulseq.calc_duration import calc_duration


def align(**kwargs: Union[SimpleNamespace, List[SimpleNamespace]]) -> List[SimpleNamespace]:
    """
    Sets delays of the objects within the block to achieve the desired alignment of the objects in the block. Aligns
    objects as per specified alignment options by setting delays of the pulse sequence events within the block. All
    previously configured delays within objects are taken into account during calculating of the block duration but
    then reset according to the selected alignment. Possible values for align_spec are 'left', 'center', 'right'.

    Parameters
    ----------
    args : dict{str, [SimpleNamespace, ...]}
        Dictionary mapping of alignment options and `SimpleNamespace` objects.
        Format: alignment_spec1=SimpleNamespace, alignment_spec2=[SimpleNamespace, ...], ...
        Alignment spec must be one of `left`, `center` or `right`.

    Returns
    -------
    objects : [SimpleNamespace, ...]
        List of aligned `SimpleNamespace` objects.

    Raises
    ------
    ValueError
        If first parameter is not of type `str`.
        If invalid alignment spec is passed. Must be one of `left`, `center` or `right`.

    Examples
    --------
    al_grad1, al_grad2, al_grad3 = align(right=[grad1, grad2, grad3])
    """
    alignment_specs = list(kwargs.keys())
    if not isinstance(alignment_specs[0], str):
        raise ValueError(f'First parameter must be of type str. Passed: {type(alignment_specs[0])}')

    alignment_options = ['left', 'center', 'right']
    if np.any([align_opt not in alignment_options for align_opt in alignment_specs]):
        raise ValueError('Invalid alignment spec.')

    alignments = []
    objects = []
    for curr_align in alignment_specs:
        objects_to_align = kwargs[curr_align]
        curr_align = alignment_options.index(curr_align)
        if isinstance(objects_to_align, (list, np.ndarray, tuple)):
            alignments.extend([curr_align] * len(objects_to_align))
            objects.extend(objects_to_align)
        elif isinstance(objects_to_align, SimpleNamespace):
            alignments.extend([curr_align])
            objects.append(objects_to_align)

    dur = calc_duration(*objects)

    # copy() to emulate pass-by-value; otherwise passed events are modified
    objects = deepcopy(objects)

    # Set new delays
    for i in range(len(objects)):
        if alignments[i] == 0:
            objects[i].delay = 0
        elif alignments[i] == 1:
            objects[i].delay = (dur - calc_duration(objects[i])) / 2
        elif alignments[i] == 2:
            objects[i].delay = dur - calc_duration(objects[i]) + objects[i].delay
            if objects[i].delay < 0:
                raise ValueError(
                    'align() attempts to set a negative delay, probably some RF pulses ignore rf_ringdown_time'
                )

    return objects



"""Tests for the make_adiabatic_pulse.py module

Will Clarke, University of Oxford, 2024
"""

import itertools

import numpy as np
import pytest
from pypulseq import make_adiabatic_pulse
from pypulseq.supported_labels_rf_use import get_supported_rf_uses


def test_pulse_select():
    valid_rf_use_labels = get_supported_rf_uses()
    valid_pulse_types = ('hypsec', 'wurst')

    # Check all use and valid pulse combinations return a sensible object
    # with default parameters.
    for pulse_type, use_label in itertools.product(valid_pulse_types, valid_rf_use_labels):
        rf_obj = make_adiabatic_pulse(pulse_type=pulse_type, use=use_label)
        assert rf_obj.type == 'rf'
        assert rf_obj.use == use_label

    # Check the appropriate errors are raised if we specify nonsense
    with pytest.raises(ValueError, match=r'Invalid type parameter\. Must be one of '):
        make_adiabatic_pulse(pulse_type='not a pulse type')

    with pytest.raises(ValueError, match=r'Invalid type parameter\. Must be one of '):
        make_adiabatic_pulse(pulse_type='')

    with pytest.raises(ValueError, match=r'Invalid use parameter\. Must be one of '):
        make_adiabatic_pulse(pulse_type='hypsec', use='not a use')

    # Default use case
    rf_obj = make_adiabatic_pulse(pulse_type='hypsec')
    assert rf_obj.use == 'inversion'


def test_option_requirements():
    # Require non-zero slice thickness if grad requested
    with pytest.raises(ValueError, match='Slice thickness must be provided'):
        make_adiabatic_pulse(pulse_type='hypsec', return_gz=True)

    _, gz, gzr = make_adiabatic_pulse(pulse_type='hypsec', return_gz=True, slice_thickness=1)
    assert gz.type == 'trap'
    assert gzr.type == 'trap'


# My intention was to test that the rephase gradient area is appropriate,
# but this doesn't pass and I'm highly suspicious of the calculation in
# the code
def test_returned_grads():
    pass
    # _, gz, gzr = make_adiabatic_pulse(
    #         pulse_type="hypsec",
    #         return_gz=True,
    #         slice_thickness=1)
    # assert np.isclose(-gz.area / 2, gzr.area)


def test_hypsec_options():
    pobj = make_adiabatic_pulse(pulse_type='hypsec', beta=700, mu=6, duration=0.05)

    assert np.isclose(pobj.shape_dur, 0.05)


def test_wurst_options():
    pobj = make_adiabatic_pulse(pulse_type='wurst', n_fac=25, bandwidth=30000, duration=0.05)

    assert np.isclose(pobj.shape_dur, 0.05)



from types import SimpleNamespace
from typing import Tuple, Union

import numpy as np

from pypulseq import eps
from pypulseq.make_extended_trapezoid import make_extended_trapezoid
from pypulseq.opts import Opts
from pypulseq.utils.cumsum import cumsum
from pypulseq.utils.tracing import trace, trace_enabled


def make_extended_trapezoid_area(
    area: float,
    channel: str,
    grad_start: float,
    grad_end: float,
    convert_to_arbitrary: bool = False,
    system: Union[Opts, None] = None,
) -> Tuple[SimpleNamespace, np.array, np.array]:
    """Make the shortest possible extended trapezoid for given area and gradient start and end point.

    Parameters
    ----------
    area : float
        Area of extended trapezoid.
    channel : str
        Orientation of extended trapezoidal gradient event. Must be one of 'x', 'y' or 'z'.
    grad_start : float
        Starting non-zero gradient value.
    grad_end : float
        Ending non-zero gradient value.
    convert_to_arbitrary : bool, default=False
        Boolean flag to enable converting the extended trapezoid gradient into an arbitrary gradient.
    system: Opts, optional
        System limits.

    Returns
    -------
    grad : SimpleNamespace
        Extended trapezoid event.
    times : numpy.ndarray
        Time points of the extended trapezoid.
    amplitude : numpy.ndarray
        Amplitude values of the extended trapezoid.

    Raises
    ------
        ValueError if no solution was found that satisfies the constraints and the desired area.
    """
    if system is None:
        system = Opts.default

    max_slew = system.max_slew * 0.99
    max_grad = system.max_grad * 0.99
    raster_time = system.grad_raster_time

    def _to_raster(time: float) -> float:
        return np.ceil(time / raster_time) * raster_time

    def _calc_ramp_time(grad_1: float, grad_2: float) -> float:
        return _to_raster(abs(grad_1 - grad_2) / max_slew)

    def _find_solution(duration: int) -> Union[None, Tuple[int, int, int, float]]:
        """Find extended trapezoid gradient waveform for given duration.

        The function performs a grid search over all possible ramp-up, ramp-down and flat times
        for the given duration and returns the solution with the lowest slew rate.

        Parameters
        ----------
        duration
            duration of the gradient in integer multiples of raster_time

        Returns
        -------
            Tuple of ramp-up time, flat time, ramp-down time, gradient amplitude or None if no solution was found
        """
        # Determine timings to check for possible solutions
        ramp_up_times = []
        ramp_down_times = []

        # First, consider solutions that use maximum slew rate:
        # Analytically calculate calculate the point where:
        #   grad_start + ramp_up_time * max_slew == grad_end + ramp_down_time * max_slew
        ramp_up_time = (duration * max_slew * raster_time - grad_start + grad_end) / (2 * max_slew * raster_time)
        ramp_up_time = round(ramp_up_time)

        # Check if gradient amplitude exceeds max_grad, if so, adjust ramp
        # times for a trapezoidal gradient with maximum slew rate.
        if grad_start + ramp_up_time * max_slew * raster_time > max_grad + eps:
            ramp_up_time = round(_calc_ramp_time(grad_start, max_grad) / raster_time)
            ramp_down_time = round(_calc_ramp_time(grad_end, max_grad) / raster_time)
        else:
            ramp_down_time = duration - ramp_up_time

        # Add possible solution if timing is valid
        if ramp_up_time > 0 and ramp_down_time > 0 and ramp_up_time + ramp_down_time <= duration:
            ramp_up_times.append(ramp_up_time)
            ramp_down_times.append(ramp_down_time)

        # Analytically calculate calculate the point where:
        #   grad_start - ramp_up_time * max_slew == grad_end - ramp_down_time * max_slew
        ramp_up_time = (duration * max_slew * raster_time + grad_start - grad_end) / (2 * max_slew * raster_time)
        ramp_up_time = round(ramp_up_time)

        # Check if gradient amplitude exceeds -max_grad, if so, adjust ramp
        # times for a trapezoidal gradient with maximum slew rate.
        if grad_start - ramp_up_time * max_slew * raster_time < -max_grad - eps:
            ramp_up_time = round(_calc_ramp_time(grad_start, -max_grad) / raster_time)
            ramp_down_time = round(_calc_ramp_time(grad_end, -max_grad) / raster_time)
        else:
            ramp_down_time = duration - ramp_up_time

        # Add possible solution if timing is valid
        if ramp_up_time > 0 and ramp_down_time > 0 and ramp_up_time + ramp_down_time <= duration:
            ramp_up_times.append(ramp_up_time)
            ramp_down_times.append(ramp_down_time)

        # Second, try any solution with flat_time == 0
        # This appears to be necessary for many cases, but going through all
        # timings here is probably too conservative still.
        for ramp_up_time in range(1, duration):
            ramp_up_times.append(ramp_up_time)
            ramp_down_times.append(duration - ramp_up_time)

        time_ramp_up = np.array(ramp_up_times)
        time_ramp_down = np.array(ramp_down_times)

        # Calculate corresponding flat times
        flat_time = duration - time_ramp_up - time_ramp_down

        # Filter search space for valid timings (flat time >= 0)
        valid_indices = flat_time >= 0
        time_ramp_up = time_ramp_up[valid_indices]
        time_ramp_down = time_ramp_down[valid_indices]
        flat_time = flat_time[valid_indices]

        # Calculate gradient strength for given timing using analytical solution
        grad_amp = -(time_ramp_up * raster_time * grad_start + time_ramp_down * raster_time * grad_end - 2 * area) / (
            (time_ramp_up + 2 * flat_time + time_ramp_down) * raster_time
        )

        # Calculate slew rates for given timings
        slew_rate1 = abs(grad_start - grad_amp) / (time_ramp_up * raster_time)
        slew_rate2 = abs(grad_end - grad_amp) / (time_ramp_down * raster_time)

        # Filter solutions that satisfy max_grad and max_slew constraints
        valid_indices = (
            (abs(grad_amp) <= max_grad + 1e-8) & (slew_rate1 <= max_slew + 1e-8) & (slew_rate2 <= max_slew + 1e-8)
        )
        solutions = np.flatnonzero(valid_indices)

        # Check if any valid solutions were found
        if solutions.shape[0] == 0:
            return None

        # Find solution with lowest slew rate and return it
        ind = np.argmin(slew_rate1[valid_indices] + slew_rate2[valid_indices])
        ind = solutions[ind]
        return (int(time_ramp_up[ind]), int(flat_time[ind]), int(time_ramp_down[ind]), float(grad_amp[ind]))

    # Perform a linear search
    # This is necessary because there can exist a dead space where solutions
    # do not exist for some durations longer than the optimal duration. The
    # binary search below fails to find the optimum in those cases.
    # TODO: Check if range is sufficient, try to calculate the dead space.
    min_duration = max(round(_calc_ramp_time(grad_end, grad_start) / raster_time), 2)

    # Calculate duration needed to ramp down gradient to zero.
    # From this point onwards, solutions can always be found by extending
    # the duration and doing a binary search.
    max_duration = max(
        round(_calc_ramp_time(0, grad_start) / raster_time),
        round(_calc_ramp_time(0, grad_end) / raster_time),
        min_duration,
    )

    # Linear search
    solution = None
    for duration in range(min_duration, max_duration + 1):
        solution = _find_solution(duration)
        if solution:
            break

    # Perform a binary search for duration > max_duration if no solution was found
    if not solution:
        # First, find the upper limit on duration where a solution exists by
        # exponentially expanding the duration.
        while not solution:
            max_duration *= 2
            solution = _find_solution(max_duration)

        def binary_search(fun, lower_limit, upper_limit):
            if lower_limit == upper_limit - 1:
                return fun(upper_limit)

            test_value = (upper_limit + lower_limit) // 2

            if fun(test_value):
                return binary_search(fun, lower_limit, test_value)
            else:
                return binary_search(fun, test_value, upper_limit)

        solution = binary_search(_find_solution, max_duration // 2, max_duration)

    # Get timing and gradient amplitude from solution
    time_ramp_up = solution[0] * raster_time
    flat_time = solution[1] * raster_time
    time_ramp_down = solution[2] * raster_time
    grad_amp = solution[3]

    # Create extended trapezoid
    if flat_time > 0:
        times = cumsum(0, time_ramp_up, flat_time, time_ramp_down)
        amplitudes = np.array([grad_start, grad_amp, grad_amp, grad_end])
    else:
        times = cumsum(0, time_ramp_up, time_ramp_down)
        amplitudes = np.array([grad_start, grad_amp, grad_end])

    grad = make_extended_trapezoid(
        channel=channel, amplitudes=amplitudes, convert_to_arbitrary=convert_to_arbitrary, system=system, times=times
    )

    # Overwrite trace
    if trace_enabled():
        grad.trace = trace()

    if not abs(grad.area - area) < 1e-8:
        raise ValueError(f'Could not find a solution for area={area}.')

    return grad, grad.tt, grad.waveform



import math
import warnings
from types import SimpleNamespace
from typing import Literal, Union

from pypulseq import eps
from pypulseq.opts import Opts
from pypulseq.utils.tracing import trace, trace_enabled


def calculate_shortest_params_for_area(area: float, max_slew: float, max_grad: float, grad_raster_time: float):
    """Calculate the shortest possible rise_time, flat_time, and fall_time for a given area."""

    # Calculate initial rise time constrained by max slew rate
    rise_time = math.ceil(math.sqrt(abs(area) / max_slew) / grad_raster_time) * grad_raster_time
    rise_time = max(rise_time, grad_raster_time)

    # Calculate initial amplitude
    amplitude = area / rise_time
    effective_time = rise_time

    # Adjust for max gradient constraint
    if abs(amplitude) > max_grad + eps:
        effective_time = math.ceil(abs(area) / max_grad / grad_raster_time) * grad_raster_time
        amplitude = area / effective_time
        rise_time = math.ceil(abs(amplitude) / max_slew / grad_raster_time) * grad_raster_time
        rise_time = max(rise_time, grad_raster_time)

    # Calculate flat and fall times
    flat_time = effective_time - rise_time
    fall_time = rise_time

    return amplitude, rise_time, flat_time, fall_time


def calculate_shortest_rise_time(amplitude: float, max_slew: float, grad_raster_time: float):
    """Calculate the shortest possible rise / fall time for a given amplitude and slew rate."""

    return math.ceil(max(abs(amplitude) / max_slew, grad_raster_time) / grad_raster_time) * grad_raster_time


def make_trapezoid(
    channel: str,
    amplitude: Union[float, None] = None,
    area: Union[float, None] = None,
    delay: float = 0.0,
    duration: Union[float, None] = None,
    fall_time: Union[float, None] = None,
    flat_area: Union[float, None] = None,
    flat_time: Union[float, None] = None,
    max_grad: Union[float, None] = None,
    max_slew: Union[float, None] = None,
    rise_time: Union[float, None] = None,
    system: Union[Opts, None] = None,
) -> SimpleNamespace:
    """
    Create a trapezoidal gradient event.

    The user must supply any of the following sets of parameters:
    Area based:
    - area
    - area and duration
    - area and duration and rise_time
    - flat_time, area and rise_time
    Amplitude based:
    - amplitude and duration
    - amplitude and flat_time
    Flat area based:
    - flat_area and flat_time
    Additional options may be supplied with the above.

    See Also
    --------
    - `pypulseq.Sequence.sequence.Sequence.add_block()`
    - `pypulseq.opts.Opts`

    Parameters
    ----------
    channel : str
        Orientation of trapezoidal gradient event. Must be one of `x`, `y` or `z`.
    amplitude : float, default=None
        Peak amplitude (Hz/m).
    area : float, default=None
        Area (1/m).
    delay : float, default=0
        Delay in seconds (s).
    duration : float, default=None
        Duration in seconds (s). Duration is defined as rise_time + flat_time + fall_time.
    fall_time : float, default=None
        Fall time in seconds (s).
    flat_area : float, default=None
        Flat area (1/m).
    flat_time : float, default=None
        Flat duration in seconds (s). Default is -1 to allow for triangular pulses.
    max_grad : float, default=None
        Maximum gradient strength (Hz/m).
    max_slew : float, default=None
        Maximum slew rate (Hz/m/s).
    rise_time : float, default=0
        Rise time in seconds (s).
    system : Opts, default=Opts()
        System limits.

    Returns
    -------
    grad : SimpleNamespace
        Trapezoidal gradient event created based on the supplied parameters.

    Raises
    ------
    ValueError
        If none of `area`, `flat_area` and `amplitude` are passed
        If requested area is too large for this gradient
        If `flat_time`, `duration` and `area` are not supplied.
        Amplitude violation
    NotImplementedError
        If an input set that might be valid but not implemented is passed.
    """
    if system is None:
        system = Opts.default

    if channel not in ['x', 'y', 'z']:
        raise ValueError(f'Invalid channel. Must be one of `x`, `y` or `z`. Passed: {channel}')

    if max_grad is None:
        max_grad = system.max_grad

    if max_slew is None:
        max_slew = system.max_slew

    # If either of rise_time or fall_time is not provided, set it to the other.
    rise_time = rise_time or fall_time
    fall_time = fall_time or rise_time

    # Check which one of area, flat_area and amplitude is provided, and determine the calculation path accordingly.
    calc_path: Literal['area', 'flat_area', 'amplitude']
    if area is not None and flat_area is None and amplitude is None:
        calc_path = 'area'
    elif area is None and flat_area is not None and amplitude is None:
        calc_path = 'flat_area'
    elif area is None and flat_area is None and amplitude is not None:
        calc_path = 'amplitude'
    elif area is None and flat_area is not None and amplitude is not None:
        raise NotImplementedError('Flat Area + Amplitude input pair is not implemented yet.')
    elif area is not None and flat_area is None and amplitude is not None:
        raise NotImplementedError('Amplitude + Area input pair is not implemented yet.')
    else:
        raise ValueError("Must supply either 'area', 'flat_area' or 'amplitude'.")

    # Check if sufficient timing parameters are provided.
    if flat_time is not None and flat_area is None and amplitude is None and (rise_time is None or area is None):
        raise ValueError(
            'When `flat_time` is provided, either `flat_area`, '
            'or `amplitude`, or `rise_time` and `area` must be provided as well.'
        )

    if calc_path == 'area':
        # We have area and duration.
        if duration is not None and flat_time is None:
            if rise_time is None:
                _, rise_time, flat_time, fall_time = calculate_shortest_params_for_area(
                    area, max_slew, max_grad, system.grad_raster_time
                )
                min_duration = rise_time + flat_time + fall_time
                assert duration >= min_duration, (
                    f'Requested area is too large for this gradient. Minimum required duration is '
                    f'{round(min_duration * 1e6)} us'
                )

                dc = 1 / abs(2 * max_slew) + 1 / abs(2 * max_slew)
                amplitude2 = (duration - math.sqrt(duration**2 - 4 * abs(area) * dc)) / (2 * dc)
            else:
                if duration <= (rise_time + eps):
                    raise ValueError('The `duration` is too short for the given `rise_time`.')

                if fall_time is None:
                    fall_time = rise_time

                amplitude2 = area / (duration - 0.5 * rise_time - 0.5 * fall_time)
                possible = duration >= (rise_time + fall_time) and abs(amplitude2) <= max_grad
                assert possible, (
                    f'Requested area is too large for this gradient. Probably amplitude is violated '
                    f'{round(abs(amplitude2) / max_grad * 100)}'
                )
            flat_time = duration - rise_time - fall_time
            amplitude2 = area / (rise_time / 2 + fall_time / 2 + flat_time)

        elif flat_time is not None:
            if rise_time is None:
                raise ValueError('Must supply `rise_time` when `area` and `flat_time` is provided.')

            amplitude2 = area / (rise_time + flat_time)

        else:
            if rise_time is not None or fall_time is not None:
                warnings.warn('Rise time and fall time is ignored when calculating the shortest duration from `area`.')

            amplitude2, rise_time, flat_time, fall_time = calculate_shortest_params_for_area(
                area, max_slew, max_grad, system.grad_raster_time
            )

    elif calc_path == 'flat_area':
        # Check and raise invalid input sets
        if duration is not None:
            raise NotImplementedError('Flat Area + Duration input pair is not implemented yet.')
        elif flat_time is not None:
            amplitude2 = flat_area / flat_time

    elif calc_path == 'amplitude':
        if rise_time is None:
            rise_time = abs(amplitude) / max_slew
            rise_time = math.ceil(rise_time / system.grad_raster_time) * system.grad_raster_time
            if rise_time == 0:
                rise_time = system.grad_raster_time
            fall_time = rise_time

        amplitude2 = amplitude
        if duration is not None and flat_time is None:
            flat_time = duration - rise_time - fall_time
        elif flat_time is not None and duration is None:
            pass
        else:
            raise ValueError('Must supply area or duration.')

    if rise_time is None and fall_time is None:
        rise_time = fall_time = calculate_shortest_rise_time(amplitude2, max_slew, system.grad_raster_time)

    if abs(amplitude2) > max_grad + eps:
        raise ValueError(f'Refined amplitude ({abs(amplitude2):0.0f} Hz/m) is larger than max ({max_grad:0.0f} Hz/m).')

    if abs(amplitude2) / rise_time > max_slew * (1 + eps):
        raise ValueError(
            f'Refined slew rate ({abs(amplitude2) / rise_time:0.0f} Hz/m/s) for ramp up is larger than max ({max_slew:0.0f} Hz/m/s).'
        )

    if abs(amplitude2) / fall_time > max_slew * (1 + eps):
        raise ValueError(
            f'Refined slew rate ({abs(amplitude2) / fall_time:0.0f} Hz/m/s) for ramp down is larger than max ({max_slew:0.0f} Hz/m/s).'
        )

    grad = SimpleNamespace()
    grad.type = 'trap'
    grad.channel = channel
    grad.amplitude = amplitude2
    grad.rise_time = rise_time
    grad.flat_time = flat_time
    grad.fall_time = fall_time
    grad.area = amplitude2 * (flat_time + rise_time / 2 + fall_time / 2)
    grad.flat_area = amplitude2 * flat_time
    grad.delay = delay
    grad.first = 0
    grad.last = 0

    if trace_enabled():
        grad.trace = trace()

    return grad



"""Tests for the make_trapezoid module"""

import pytest
from pypulseq import Opts, eps, make_trapezoid


def test_channel_error():
    with pytest.raises(ValueError, match=r'Invalid channel. Must be one of `x`, `y` or `z`. Passed:'):
        make_trapezoid(channel='p')


def test_area_flatarea_amplitude_error():
    with pytest.raises(ValueError, match=r"Must supply either 'area', 'flat_area' or 'amplitude'."):
        make_trapezoid(channel='x')


def test_flat_time_error():
    errstr = (
        'When `flat_time` is provided, either `flat_area`, '
        'or `amplitude`, or `rise_time` and `area` must be provided as well.'
    )

    with pytest.raises(ValueError, match=errstr):
        make_trapezoid(channel='x', flat_time=10, area=10)


def test_area_too_large_error():
    errstr = 'Requested area is too large for this gradient. Minimum required duration is'

    with pytest.raises(AssertionError, match=errstr):
        make_trapezoid(channel='x', area=1e6, duration=1e-6)


def test_area_too_large_error_rise_time():
    errstr = 'Requested area is too large for this gradient. Probably amplitude is violated'

    with pytest.raises(AssertionError, match=errstr):
        make_trapezoid(channel='x', area=1e6, duration=1e-6, rise_time=1e-7)


def test_no_area_no_duration_error():
    errstr = 'Must supply area or duration.'

    with pytest.raises(ValueError, match=errstr):
        make_trapezoid(channel='x', amplitude=1)


def test_amplitude_too_large_error():
    errstr = r'Refined amplitude \(\d+ Hz/m\) is larger than max \(\d+ Hz/m\).'

    with pytest.raises(ValueError, match=errstr):
        make_trapezoid(channel='x', amplitude=1e10, duration=1)


def test_duration_too_short_error():
    errstr = 'The `duration` is too short for the given `rise_time`.'

    with pytest.raises(ValueError, match=errstr):
        make_trapezoid(channel='x', area=1, duration=0.1, rise_time=0.1)


def test_notimplemented_input_pairs():
    # flat_area + duration
    with pytest.raises(NotImplementedError, match=r'Flat Area \+ Duration input pair is not implemented yet.'):
        make_trapezoid(channel='x', flat_area=1, duration=1)
    # flat_area + amplitude
    with pytest.raises(NotImplementedError, match=r'Flat Area \+ Amplitude input pair is not implemented yet.'):
        make_trapezoid(channel='x', flat_area=1, amplitude=1)
    # area + amplitude
    with pytest.raises(NotImplementedError, match=r'Amplitude \+ Area input pair is not implemented yet.'):
        make_trapezoid(channel='x', area=1, amplitude=1)
        # compare_trap_out(trap, 1, 2e-5, 0, 2e-5)


def round2raster(x, raster=1e-5):
    return round(x / raster) * raster


def compare_trap_out(trap, amplitude, rise_time, flat_time, fall_time):
    assert abs(trap.amplitude - amplitude) < eps
    assert abs(trap.rise_time - rise_time) < eps
    assert abs(trap.flat_time - flat_time) < eps
    assert abs(trap.fall_time - fall_time) < eps


def test_generation_methods():
    """Test minimum input cases
    Cover:
        - area
        - amplitude and duration
        - flat_time and flat_area
        - flat_time and amplitude
        - flat_time, area and rise_time
    """

    opts = Opts()
    # Amplitude specified
    # amplitude + duration
    trap = make_trapezoid(channel='x', amplitude=1, duration=1)
    compare_trap_out(trap, 1, 1e-5, 1 - 2e-5, 1e-5)

    # flat_time + amplitude
    trap = make_trapezoid(channel='x', flat_time=1, amplitude=1)
    compare_trap_out(trap, 1, 1e-5, 1, 1e-5)

    # Flat area specified
    # flat_area + flat_time
    trap = make_trapezoid(channel='x', flat_time=1, flat_area=1)
    compare_trap_out(trap, 1, 1e-5, 1, 1e-5)

    # Area specified
    # area
    # triangle case
    trap = make_trapezoid(channel='x', area=1)
    compare_trap_out(trap, 50000, 2e-5, 0, 2e-5)
    # trap case
    trap = make_trapezoid(channel='x', area=opts.max_grad * 2)
    time_to_max = round2raster(opts.max_grad / opts.max_slew)
    compare_trap_out(
        trap, opts.max_grad, time_to_max, (opts.max_grad * 2 - time_to_max * opts.max_grad) / opts.max_grad, time_to_max
    )

    # area + duration
    trap = make_trapezoid(channel='x', area=1, duration=1)
    compare_trap_out(trap, 1.00002, 2e-5, 1 - 4e-5, 2e-5)

    # area + duration + rise_time
    trap = make_trapezoid(channel='x', area=1, duration=1, rise_time=0.01)
    compare_trap_out(trap, 1 / 0.99, 0.01, 0.98, 0.01)
    # flat_time + area + rise_time
    trap = make_trapezoid(channel='x', flat_time=0.5, area=1, rise_time=0.1)
    compare_trap_out(trap, 1 / 0.6, 0.1, 0.5, 0.1)



from itertools import combinations_with_replacement

import pypulseq as pp
import pytest

"""
Tests calc_duration by feeding it some sample events with known durations.
Additionally, tests the combination of any 2 and 3 of those events.
"""


def test_no_event():
    assert pp.calc_duration() == 0


known_duration_event_zoo = [
    ('trapz_amp1', pp.make_trapezoid('x', amplitude=1, duration=1), 1),
    ('trapz_amp1_delayed1', pp.make_trapezoid('x', amplitude=1, duration=1, delay=1), 2),
    ('delay1', pp.make_delay(1), 1),
    ('delay0', pp.make_delay(0), 0),
    ('rf0_block1', pp.make_block_pulse(flip_angle=0, duration=1), 1),
    ('rf10_block1', pp.make_block_pulse(flip_angle=10, duration=1), 1),
    ('rf10_block1_delay1', pp.make_block_pulse(flip_angle=10, duration=1, delay=1), 2),
    ('adc3', pp.make_adc(duration=3, num_samples=1), 3),
    ('adc3_delayed', pp.make_adc(duration=3, delay=1, num_samples=1), 4),
    ('outputOsc042', pp.make_digital_output_pulse('osc0', duration=42), 42),
    ('outputOsc142_delay3', pp.make_digital_output_pulse('osc1', duration=42, delay=1), 43),
    ('outputExt42_delay9', pp.make_digital_output_pulse('osc1', duration=42, delay=9), 51),
    ('triggerPhysio159', pp.make_trigger('physio1', duration=59), 59),
    ('triggerPhysio259_delay1', pp.make_trigger('physio2', duration=59, delay=1), 60),
    ('label0', pp.make_label(label='SLC', type='SET', value=0), 0),
]


@pytest.mark.parametrize('name,event,expected_dur', known_duration_event_zoo)
def test_single_events(name, event, expected_dur):
    assert pp.calc_duration(event) == expected_dur


def known_duration_event_zoo_combos(num_to_combine):
    for combo in combinations_with_replacement(known_duration_event_zoo, num_to_combine):
        name = ','.join(event[0] for event in combo)
        expected_dur = max(event[2] for event in combo)
        events = (event[1] for event in combo)
        yield name, tuple(events), expected_dur


@pytest.mark.parametrize('name,events,expected_total_dur', known_duration_event_zoo_combos(2))
def test_event_combinations2(name, events, expected_total_dur):
    assert pp.calc_duration(*events) == expected_total_dur


@pytest.mark.parametrize('name,events,expected_total_dur', known_duration_event_zoo_combos(3))
def test_event_combinations3(name, events, expected_total_dur):
    assert pp.calc_duration(*events) == expected_total_dur



class SigpyPulseOpts:
    def __init__(
        self,
        pulse_type: str = 'slr',
        ptype: str = 'st',
        ftype: str = 'ls',
        d1: float = 0.01,
        d2: float = 0.01,
        cancel_alpha_phs: bool = False,
        n_bands: int = 3,
        band_sep: int = 20,
        phs_0_pt: str = 'None',
    ):
        self.pulse_type = pulse_type

        if pulse_type == 'slr':
            self.ptype = ptype
            self.ftype = ftype
            self.d1 = d1
            self.d2 = d2
            self.cancel_alpha_phs = cancel_alpha_phs

        if pulse_type == 'sms':
            self.ptype = ptype
            self.ftype = ftype
            self.d1 = d1
            self.d2 = d2
            self.cancel_alpha_phs = cancel_alpha_phs
            self.n_bands = n_bands
            self.band_sep = band_sep
            self.phs_0_pt = phs_0_pt

    def __str__(self) -> str:
        s = 'Pulse options:'
        s += '\nptype: ' + str(self.ptype)
        s += '\nftype: ' + str(self.ftype)
        s += '\nd1: ' + str(self.d1)
        s += '\nd2: ' + str(self.d2)
        s += '\ncancel_alpha_phs: ' + str(self.cancel_alpha_phs)

        return s



# inserted for trigger support by mveldmann

from types import SimpleNamespace
from typing import Union

from pypulseq.opts import Opts


def make_trigger(
    channel: str,
    delay: float = 0.0,
    duration: float = 0.0,
    system: Union[Opts, None] = None,
) -> SimpleNamespace:
    """
     Create a trigger halt event for a synchronization with an external signal from a given channel with an optional
     given delay prio to the sync and duration after the sync. Possible channel values: 'physio1','physio2'
     (Siemens specific).

    See also `pypulseq.Sequence.sequence.Sequence.add_block()`.

    Parameters
    ----------
     channel : str
         Must be one of 'physio1' or 'physio2'.
     delay : float, default=0
         Delay in seconds
     duration: float, default=0
         Duration in seconds.
     system : Opts, default=Opts()
         System limits.

    Returns
    -------
     trigger : SimpleNamespace
         Trigger event.

    Raises
    ------
     ValueError
         If invalid `channel` is passed. Must be one of 'physio1' or 'physio2'.
    """
    if system is None:
        system = Opts.default

    if channel not in ['physio1', 'physio2']:
        raise ValueError(f"Channel {channel} is invalid. Must be one of 'physio1' or 'physio2'.")

    trigger = SimpleNamespace()
    trigger.type = 'trigger'
    trigger.channel = channel
    trigger.delay = delay
    trigger.duration = duration
    if trigger.duration <= system.grad_raster_time:
        trigger.duration = system.grad_raster_time

    return trigger



from types import SimpleNamespace
from typing import Tuple


def block_to_events(*args: SimpleNamespace | float) -> Tuple[SimpleNamespace, ...]:
    """
    Converts `args` from a block to a list of events. If `args` is already a list of event(s), returns it unmodified.

    Parameters
    ----------
    args : SimpleNamespace
        Block to be flattened into a list of events.

    Returns
    -------
    events : list[SimpleNamespace]
        List of events comprising `args` if it was a block, otherwise `args` unmodified.
    """
    if len(args) == 1 and hasattr(args[0], 'rf'):
        events = list(vars(args[0]).values())  # Get all attrs
        events = list(filter(lambda filter_none: filter_none is not None, events))  # Filter None attributes
        events = __get_label_events_if_any(*events)  # Flatten label events from dict datatype
        events = tuple(events)

    else:  # args is a tuple of events
        return args

    return events


def __get_label_events_if_any(*events: list) -> list:
    # Are any of the events labels? If yes, extract them from dict()
    final_events = []
    for e in events:
        if isinstance(e, dict):  # Only labels are stored as dicts
            final_events.extend(e.values())
        else:
            final_events.append(e)

    return final_events



import math
from copy import copy
from types import SimpleNamespace
from typing import Tuple, Union
from warnings import warn

import numpy as np

from pypulseq.make_trapezoid import make_trapezoid
from pypulseq.opts import Opts
from pypulseq.supported_labels_rf_use import get_supported_rf_uses
from pypulseq.utils.tracing import trace, trace_enabled


def make_gauss_pulse(
    flip_angle: float,
    apodization: float = 0.0,
    bandwidth: float = 0.0,
    center_pos: float = 0.5,
    delay: float = 0.0,
    dwell: float = 0.0,
    duration: float = 4e-3,
    freq_offset: float = 0.0,
    max_grad: float = 0.0,
    max_slew: float = 0.0,
    phase_offset: float = 0.0,
    return_gz: bool = False,
    slice_thickness: float = 0.0,
    system: Union[Opts, None] = None,
    time_bw_product: float = 4.0,
    use: str = 'undefined',
    freq_ppm: float = 0.0,
    phase_ppm: float = 0.0,
) -> Union[
    SimpleNamespace,
    Tuple[SimpleNamespace, SimpleNamespace, SimpleNamespace],
]:
    """
    Create a [optionally slice selective] Gauss pulse.

    See also `pypulseq.Sequence.sequence.Sequence.add_block()`.

    Parameters
    ----------
    flip_angle : float
        Flip angle in radians.
    apodization : float, default=0
        Apodization.
    bandwidth : float, default=0
        Bandwidth in Hertz (Hz).
    center_pos : float, default=0.5
        Position of peak.
    delay : float, default=0
        Delay in seconds (s).
    dwell : float, default=0
    duration : float, default=4e-3
        Duration in seconds (s).
    freq_offset : float, default=0
        Frequency offset in Hertz (Hz).
    max_grad : float, default=0
        Maximum gradient strength of accompanying slice select trapezoidal event.
    max_slew : float, default=0
        Maximum slew rate of accompanying slice select trapezoidal event.
    phase_offset : float, default=0
        Phase offset in Hertz (Hz).
    return_gz : bool, default=False
        Boolean flag to indicate if the slice-selective gradient has to be returned.
    slice_thickness : float, default=0
        Slice thickness of accompanying slice select trapezoidal event. The slice thickness determines the area of the
        slice select event.
    system : Opts, default=Opts()
        System limits.
    time_bw_product : float, default=4.0
        Time-bandwidth product.
    use : str, default='undefined'
        Use of radio-frequency Gauss pulse event.
        Must be one of 'excitation', 'refocusing', 'inversion',
        'saturation', 'preparation', 'other', 'undefined'.
    freq_ppm : float, default=0
        PPM frequency offset.
    phase_ppm : float, default=0
        PPM phase offset.

    Returns
    -------
    rf : SimpleNamespace
        Radio-frequency gauss pulse event.
    gz : SimpleNamespace, optional
        Accompanying slice select trapezoidal gradient event.
    gzr : SimpleNamespace, optional
        Accompanying slice select rephasing trapezoidal gradient event.

    Raises
    ------
    ValueError
        If invalid `use` is passed.
        If `return_gz=True` and `slice_thickness` was not passed.
    """
    if system is None:
        system = Opts.default

    valid_pulse_uses = get_supported_rf_uses()
    if use != '' and use not in valid_pulse_uses:
        raise ValueError(f'Invalid use parameter. Must be one of {valid_pulse_uses}. Passed: {use}')

    if dwell == 0:
        dwell = system.rf_raster_time

    if bandwidth == 0:
        bandwidth = time_bw_product / duration
    else:
        bandwidth = bandwidth
    alpha = apodization
    n_samples = round(duration / dwell)
    t = (np.arange(1, n_samples + 1) - 0.5) * dwell
    tt = t - (duration * center_pos)
    window = 1 - alpha + alpha * np.cos(2 * np.pi * tt / duration)
    signal = window * __gauss(bandwidth * tt)
    flip = np.sum(signal) * dwell * 2 * np.pi
    signal = signal * flip_angle / flip

    rf = SimpleNamespace()
    rf.type = 'rf'
    rf.signal = signal
    rf.t = t
    rf.shape_dur = n_samples * dwell
    rf.freq_offset = freq_offset
    rf.phase_offset = phase_offset
    rf.freq_ppm = freq_ppm
    rf.phase_ppm = phase_ppm
    rf.dead_time = system.rf_dead_time
    rf.ringdown_time = system.rf_ringdown_time
    rf.delay = delay
    rf.center = duration * center_pos
    if use != '':
        rf.use = use

    if rf.dead_time > rf.delay:
        warn(
            f'Specified RF delay {rf.delay * 1e6:.2f} us is less than the dead time {rf.dead_time * 1e6:.0f} us. Delay was increased to the dead time.',
            stacklevel=2,
        )
        rf.delay = rf.dead_time

    if return_gz:
        if slice_thickness == 0:
            raise ValueError('Slice thickness must be provided')

        if max_grad > 0:
            system = copy(system)
            system.max_grad = max_grad

        if max_slew > 0:
            system = copy(system)
            system.max_slew = max_slew

        amplitude = bandwidth / slice_thickness
        area = amplitude * duration
        gz = make_trapezoid(channel='z', system=system, flat_time=duration, flat_area=area)
        gzr = make_trapezoid(
            channel='z',
            system=system,
            area=-area * (1 - center_pos) - 0.5 * (gz.area - area),
        )

        if rf.delay > gz.rise_time:
            gz.delay = math.ceil((rf.delay - gz.rise_time) / system.grad_raster_time) * system.grad_raster_time

        if rf.delay < (gz.rise_time + gz.delay):
            rf.delay = gz.rise_time + gz.delay

    # Following 2 lines of code are workarounds for numpy returning 3.14... for np.angle(-0.00...)
    negative_zero_indices = np.where(rf.signal == -0.0)
    rf.signal[negative_zero_indices] = 0

    if trace_enabled():
        rf.trace = trace()

    if return_gz:
        return rf, gz, gzr
    else:
        return rf


def __gauss(x: np.ndarray) -> np.ndarray:
    return np.exp(-np.pi * np.square(x))



"""
Test functions for the mod_grad_axis method in pypulseq.Sequence.
"""

import numpy as np
import pypulseq as pp
import pytest


def test_mod_grad_axis_trapezoid_amplitude_scaling():
    """Test that trapezoid gradient amplitude scaling works correctly."""
    seq = pp.Sequence()

    # Create trapezoid gradients on different axes
    gx = pp.make_trapezoid('x', amplitude=1000, flat_time=5e-3)
    gy = pp.make_trapezoid('y', amplitude=2000, flat_time=3e-3)
    gz = pp.make_trapezoid('z', amplitude=1500, flat_time=4e-3)

    seq.add_block(gx, gy, gz)

    # Store original values
    original_gx_amp = gx.amplitude
    original_gx_area = gx.area
    original_gx_timing = (gx.rise_time, gx.flat_time, gx.fall_time, gx.delay)

    # Scale x-axis by factor of 2
    seq.mod_grad_axis('x', 2.0)

    # Get the modified gradient
    block = seq.get_block(1)
    modified_gx = block.gx

    # Check amplitude scaling
    assert abs(modified_gx.amplitude - 2 * original_gx_amp) < 1e-10, (
        f'Expected amplitude {2 * original_gx_amp}, got {modified_gx.amplitude}'
    )

    # Check area scaling (area should scale with amplitude)
    assert abs(modified_gx.area - 2 * original_gx_area) < 1e-10, (
        f'Expected area {2 * original_gx_area}, got {modified_gx.area}'
    )

    # Check timing unchanged
    modified_timing = (modified_gx.rise_time, modified_gx.flat_time, modified_gx.fall_time, modified_gx.delay)
    for orig, mod in zip(original_gx_timing, modified_timing, strict=False):
        assert abs(orig - mod) < 1e-12, f'Timing should not change: {original_gx_timing} vs {modified_timing}'

    # Check other axes unchanged
    assert abs(block.gy.amplitude - gy.amplitude) < 1e-12, 'Y gradient should be unchanged'
    assert abs(block.gz.amplitude - gz.amplitude) < 1e-12, 'Z gradient should be unchanged'


def test_mod_grad_axis_arbitrary_gradient_scaling():
    """Test that arbitrary gradient amplitude scaling works correctly."""
    seq = pp.Sequence()

    # Create arbitrary gradient using make_extended_trapezoid (simpler than make_arbitrary_grad)
    gx = pp.make_extended_trapezoid('x', amplitudes=[0, 1000, 1000, 0], times=[0, 1e-3, 4e-3, 5e-3])

    seq.add_block(gx)

    # Store original values
    original_amplitude = gx.waveform.max()
    original_area = np.trapezoid(gx.waveform, gx.tt)
    original_first = gx.first
    original_last = gx.last

    # Scale x-axis by factor of 3
    seq.mod_grad_axis('x', 3.0)

    # Get the modified gradient
    block = seq.get_block(1)
    modified_gx = block.gx

    # Check amplitude scaling
    assert abs(modified_gx.waveform.max() - 3 * original_amplitude) < 1e-10, 'Waveform amplitude should scale by factor'

    # Check area scaling
    modified_area = np.trapezoid(modified_gx.waveform, modified_gx.tt)
    assert abs(modified_area - 3 * original_area) < 1e-10, 'Waveform area should scale by factor'

    # Check first and last values scaling
    assert abs(modified_gx.first - 3 * original_first) < 1e-10, 'First value should scale by factor'
    assert abs(modified_gx.last - 3 * original_last) < 1e-10, 'Last value should scale by factor'

    # Check timing unchanged
    np.testing.assert_array_almost_equal(modified_gx.tt, gx.tt, decimal=12, err_msg='Time array should not change')


def test_mod_grad_axis_zero_scaling():
    """Test setting gradients to zero."""
    seq = pp.Sequence()

    # Create gradients on all axes
    gx = pp.make_trapezoid('x', amplitude=1000, flat_time=5e-3)
    gy = pp.make_trapezoid('y', amplitude=2000, flat_time=3e-3)
    gz = pp.make_trapezoid('z', amplitude=1500, flat_time=4e-3)

    seq.add_block(gx, gy, gz)

    # Zero out y-axis
    seq.mod_grad_axis('y', 0)

    # Get the modified block
    block = seq.get_block(1)

    # Check y-gradient is zeroed
    assert abs(block.gy.amplitude) < 1e-12, 'Y gradient amplitude should be zero'
    assert abs(block.gy.area) < 1e-12, 'Y gradient area should be zero'

    # Check other axes unchanged
    assert abs(block.gx.amplitude - gx.amplitude) < 1e-12, 'X gradient should be unchanged'
    assert abs(block.gz.amplitude - gz.amplitude) < 1e-12, 'Z gradient should be unchanged'

    # Check timing preserved
    assert abs(block.gy.rise_time - gy.rise_time) < 1e-12, 'Rise time should be preserved'
    assert abs(block.gy.flat_time - gy.flat_time) < 1e-12, 'Flat time should be preserved'
    assert abs(block.gy.fall_time - gy.fall_time) < 1e-12, 'Fall time should be preserved'


@pytest.mark.parametrize('factor', [2.5, -1.7, 0.3, 10.0])
def test_mod_grad_axis_consistency_scaling(factor):
    """Test consistency: scale by factor 'a' then by '1/a' should return to original."""
    seq = pp.Sequence()

    # Create gradients with different types
    gx_trap = pp.make_trapezoid('x', amplitude=1234.5, flat_time=5e-3)

    # Create arbitrary gradient using make_extended_trapezoid
    gy_arb = pp.make_extended_trapezoid('y', amplitudes=[0, 987.6, 500, 0], times=[0, 1e-3, 3e-3, 4e-3])

    gz_trap = pp.make_trapezoid('z', amplitude=-567.8, flat_time=3e-3)

    seq.add_block(gx_trap, gy_arb, gz_trap)

    # Store original values
    block_orig = seq.get_block(1)
    orig_gx_amp = block_orig.gx.amplitude
    orig_gx_area = block_orig.gx.area
    orig_gy_waveform = block_orig.gy.waveform.copy()
    orig_gy_area = np.trapezoid(block_orig.gy.waveform, block_orig.gy.tt)
    orig_gz_amp = block_orig.gz.amplitude
    orig_gz_area = block_orig.gz.area

    # Scale by factor
    seq.mod_grad_axis('x', factor)
    seq.mod_grad_axis('y', factor)
    seq.mod_grad_axis('z', factor)

    # Scale back by 1/factor
    seq.mod_grad_axis('x', 1.0 / factor)
    seq.mod_grad_axis('y', 1.0 / factor)
    seq.mod_grad_axis('z', 1.0 / factor)

    # Check we're back to original values
    block_final = seq.get_block(1)

    # Check trapezoid gradients
    assert abs(block_final.gx.amplitude - orig_gx_amp) < 1e-10, f'X amplitude not restored after factor {factor}'
    assert abs(block_final.gx.area - orig_gx_area) < 1e-10, f'X area not restored after factor {factor}'

    assert abs(block_final.gz.amplitude - orig_gz_amp) < 1e-10, f'Z amplitude not restored after factor {factor}'
    assert abs(block_final.gz.area - orig_gz_area) < 1e-10, f'Z area not restored after factor {factor}'

    # Check arbitrary gradient
    np.testing.assert_array_almost_equal(
        block_final.gy.waveform,
        orig_gy_waveform,
        decimal=10,
        err_msg=f'Y waveform not restored after factor {factor}',
    )

    final_gy_area = np.trapezoid(block_final.gy.waveform, block_final.gy.tt)
    assert abs(final_gy_area - orig_gy_area) < 1e-10, f'Y area not restored after factor {factor}'


def test_mod_grad_axis_negative_scaling():
    """Test gradient inversion with negative scaling."""
    seq = pp.Sequence()

    gx = pp.make_trapezoid('x', amplitude=1000, flat_time=5e-3)
    seq.add_block(gx)

    # Store original values
    original_amp = gx.amplitude
    original_area = gx.area

    # Invert gradient
    seq.mod_grad_axis('x', -1)

    # Check inversion
    block = seq.get_block(1)
    assert abs(block.gx.amplitude - (-original_amp)) < 1e-12, 'Amplitude should be inverted'
    assert abs(block.gx.area - (-original_area)) < 1e-12, 'Area should be inverted'


def test_mod_grad_axis_multiple_blocks():
    """Test scaling gradients across multiple blocks."""
    seq = pp.Sequence()

    # Create multiple blocks with y-gradients
    gy1 = pp.make_trapezoid('y', amplitude=1000, flat_time=2e-3)
    gy2 = pp.make_trapezoid('y', amplitude=1500, flat_time=3e-3)
    gy3 = pp.make_trapezoid('y', amplitude=-800, flat_time=1e-3)

    seq.add_block(gy1)
    seq.add_block(gy2)
    seq.add_block(gy3)

    # Scale all y-gradients by 2.5
    seq.mod_grad_axis('y', 2.5)

    # Check all blocks are scaled
    block1 = seq.get_block(1)
    block2 = seq.get_block(2)
    block3 = seq.get_block(3)

    assert abs(block1.gy.amplitude - 2.5 * 1000) < 1e-10, 'Block 1 not scaled'
    assert abs(block2.gy.amplitude - 2.5 * 1500) < 1e-10, 'Block 2 not scaled'
    assert abs(block3.gy.amplitude - 2.5 * (-800)) < 1e-10, 'Block 3 not scaled'


def test_mod_grad_axis_invalid_axis():
    """Test error handling for invalid axis."""
    seq = pp.Sequence()

    with pytest.raises(ValueError, match='Invalid axis'):
        seq.mod_grad_axis('w', 1.0)

    with pytest.raises(ValueError, match='Invalid axis'):
        seq.mod_grad_axis('xy', 1.0)


def test_mod_grad_axis_shared_gradient_error():
    """Test error when same gradient is used on multiple axes."""
    seq = pp.Sequence()

    # Create a gradient
    gx = pp.make_trapezoid('x', amplitude=1000, flat_time=5e-3)
    seq.add_block(gx)

    block_id = next(iter(seq.block_events.keys()))
    original_block = next(iter(seq.block_events.values()))

    # Modify the block to have the same gradient ID on both x and y axes
    # Block structure: [rf_id, gx_id, gy_id, gz_id, adc_id, delay_id, ...]
    modified_block = list(original_block)
    modified_block[3] = modified_block[2]  # Set gy_id = gx_id (same gradient on both axes)
    seq.block_events[block_id] = tuple(modified_block)

    # Now trying to modify x-axis should raise RuntimeError
    with pytest.raises(
        RuntimeError, match='mod_grad_axis does not yet support the same gradient event used on multiple axes'
    ):
        seq.mod_grad_axis('x', 2.0)


def test_mod_grad_axis_empty_sequence():
    """Test mod_grad_axis on empty sequence (should do nothing)."""
    seq = pp.Sequence()

    # Should not raise error
    seq.mod_grad_axis('x', 2.0)

    # Sequence should still be empty
    assert len(seq.block_events) == 0, 'Empty sequence should remain empty'


def test_mod_grad_axis_no_gradients_on_axis():
    """Test mod_grad_axis when no gradients exist on specified axis."""
    seq = pp.Sequence()

    # Add gradients only on x and z
    gx = pp.make_trapezoid('x', amplitude=1000, flat_time=5e-3)
    gz = pp.make_trapezoid('z', amplitude=1500, flat_time=3e-3)
    seq.add_block(gx, gz)

    # Try to modify y-axis (no gradients there)
    seq.mod_grad_axis('y', 2.0)  # Should do nothing

    # Check x and z are unchanged
    block = seq.get_block(1)
    assert abs(block.gx.amplitude - 1000) < 1e-12, 'X gradient should be unchanged'
    assert abs(block.gz.amplitude - 1500) < 1e-12, 'Z gradient should be unchanged'
    assert not hasattr(block, 'gy') or block.gy is None, 'Y gradient should not exist'


def test_mod_grad_axis_keymap_integrity():
    """Test that EventLibrary keymap integrity is maintained after modification."""
    seq = pp.Sequence()

    # Create a gradient and add it to the sequence
    gx = pp.make_trapezoid('x', amplitude=1000, flat_time=5e-3)
    seq.add_block(gx)

    # Store original library state
    original_keymap_size = len(seq.grad_library.keymap)
    original_data_size = len(seq.grad_library.data)

    # Modify the gradient
    seq.mod_grad_axis('x', -1.0)

    # Check that library sizes are maintained (no duplicates created)
    assert len(seq.grad_library.keymap) == original_keymap_size, 'Keymap size should be unchanged'
    assert len(seq.grad_library.data) == original_data_size, 'Data size should be unchanged'

    # Verify that a modification does not create a duplicate gradient event
    gx_inverted = pp.make_trapezoid('x', amplitude=-1000, flat_time=5e-3)
    seq.add_block(gx_inverted)

    # Library should still have the same number of unique gradients
    assert len(seq.grad_library.data) == original_data_size, 'No duplicate gradient should be created'

    # Verify that both gradients reference the same gradient event
    block1_events = seq.block_events[1]
    block2_events = seq.block_events[2]

    # The gradient event ID is at index 2
    gx_event_id_block1 = block1_events[2]
    gx_event_id_block2 = block2_events[2]

    assert gx_event_id_block1 == gx_event_id_block2, 'Both gradients should reference the same gradient event'
